#pragma once

#include <cstdint>
#include <Pokitto.h>

class SDFBitmap {
public:
    constexpr SDFBitmap() {}
    
    constexpr void set_distance_bitmap(const std::int8_t* bitmap, std::uint32_t sw, std::uint32_t sh, std::uint32_t dw, std::uint32_t dh) {
        _bitmap = bitmap;
        _bitmap_w = sw;
        _bitmap_h = sh;
        
        _w = dw;
        _h = dh;
        
        _step_u = (static_cast<std::int64_t>(sw - 1) << Q) / dw;
        _step_v = (static_cast<std::int64_t>(sh - 1) << Q) / dh;
    }
    
    constexpr std::int32_t get_distance(std::int32_t x, std::int32_t y) {
        y = y < 0 ? 0 : (y < _h ? y : (_h - 1));
        std::int32_t v0 = _step_v * y + (_step_v >> 1);
        std::int32_t vf = v0 & (ONE - 1);
        v0 >>= Q;
        
        x = x < 0 ? 0 : (x < _w ? x : (_w - 1));
        std::int32_t u0 = _step_u * x + (_step_u >> 1);
        std::int32_t uf = u0 & (ONE - 1);
        u0 >>= Q;
        
        // Fetch the two rows from the bitmap that contribute to the current line
        const std::int8_t* top_row = &_bitmap[v0 * _bitmap_w];
        const std::int8_t* bot_row = &_bitmap[(v0 + 1) * _bitmap_w];
        
        // Read the value at each of the four corners around u, v position
        std::int32_t top_l = top_row[u0];
        std::int32_t top_r = top_row[u0 + 1];
        std::int32_t bot_l = bot_row[u0];
        std::int32_t bot_r = bot_row[u0 + 1];
        
        // Use bilinear filtering to calculate distance value based on the four corner values
        std::int32_t dist = mix(mix(top_l, top_r, uf), mix(bot_l, bot_r, uf), vf);
        
        // Normalize to the range -1...1 as Q16 fixed point number
        dist *= (1 << Q) / 127;
        
        return dist;
    }

private:
    static constexpr std::uint32_t Q = 16;
    static constexpr std::int32_t ONE = 1 << Q;
    
    // Bilinear filtering
    static constexpr std::int32_t mix(std::int32_t a, std::int32_t b, std::int32_t f) { return a + ((f * (b - a)) >> Q); }

    const std::int8_t* _bitmap = nullptr;
    std::int16_t _bitmap_w = 0;
    std::int16_t _bitmap_h = 0;
    
    std::uint16_t _w = 0;
    std::uint16_t _h = 0;
    
    std::int32_t _step_u = 0;
    std::int32_t _step_v = 0;
};
